/*
 * Copyright (C) 2025 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.vorgang.redirect;

import static org.assertj.core.api.Assertions.*;
import static org.awaitility.Awaitility.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;
import static org.mockito.Mockito.eq;

import java.io.InputStream;
import java.time.Duration;
import java.util.List;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.BiFunction;
import java.util.function.Consumer;
import java.util.stream.IntStream;

import org.apache.commons.lang3.RandomUtils;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.MockedStatic;
import org.mockito.Spy;

import de.ozgcloud.common.binaryfile.GrpcFileUploadUtils;
import de.ozgcloud.common.binaryfile.StreamingFileSender;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.common.test.ReflectionTestUtils;
import de.ozgcloud.eingang.forwarder.RouteForwardingServiceGrpc;
import de.ozgcloud.eingang.forwarding.GrpcRouteForwarding;
import de.ozgcloud.eingang.forwarding.GrpcRouteForwardingRequest;
import de.ozgcloud.eingang.forwarding.GrpcRouteForwardingResponse;
import de.ozgcloud.vorgang.callcontext.VorgangManagerClientCallContextAttachingInterceptor;
import de.ozgcloud.vorgang.files.FileService;
import de.ozgcloud.vorgang.vorgang.IncomingFile;
import de.ozgcloud.vorgang.vorgang.IncomingFileGroup;
import de.ozgcloud.vorgang.vorgang.IncomingFileGroupTestFactory;
import de.ozgcloud.vorgang.vorgang.IncomingFileMapper;
import de.ozgcloud.vorgang.vorgang.IncomingFileTestFactory;
import de.ozgcloud.vorgang.vorgang.redirect.EingangForwarder.DelegatingOnReadyHandler;
import de.ozgcloud.vorgang.vorgang.redirect.EingangForwarder.ForwardingResponseObserver;
import io.grpc.stub.ClientCallStreamObserver;
import lombok.SneakyThrows;

class EingangForwarderTest {

	@Mock
	private RouteForwardingServiceGrpc.RouteForwardingServiceStub serviceStub;
	@Mock
	private FileService fileService;
	@Mock
	private IncomingFileMapper incomingFileMapper;
	@InjectMocks
	@Spy
	private EingangForwarder forwarder;

	@Nested
	class TestForward {

		@Mock
		private GrpcRouteForwarding grpcRouteForwarding;
		@Mock
		private ClientCallStreamObserver<GrpcRouteForwardingRequest> requestObserver;
		private final List<IncomingFileGroup> attachments = List.of(IncomingFileGroupTestFactory.create());
		private final List<IncomingFile> representations = List.of(IncomingFileTestFactory.create());
		@Mock
		private Future<GrpcRouteForwardingResponse> future;

		@BeforeEach
		void init() {
			doReturn(future).when(forwarder).performGrpcCall();
			doNothing().when(forwarder).sendRouteForwarding(any());
			doNothing().when(forwarder).sendAttachments(any());
			doNothing().when(forwarder).sendRepresentations(any());
			doNothing().when(forwarder).waitForCompletion(any());
			setRequestObserverInForwarder(requestObserver);
		}

		@Test
		void shouldPerformGrpcCall() {
			forwarder.forward(grpcRouteForwarding, attachments, representations);

			verify(forwarder).performGrpcCall();
		}

		@Test
		void shouldSendRouteForwarding() {
			forwarder.forward(grpcRouteForwarding, attachments, representations);

			verify(forwarder).sendRouteForwarding(grpcRouteForwarding);
		}

		@Test
		void shouldSendAttachments() {
			forwarder.forward(grpcRouteForwarding, attachments, representations);

			verify(forwarder).sendAttachments(attachments);
		}

		@Test
		void shouldSendRepresentations() {
			forwarder.forward(grpcRouteForwarding, attachments, representations);

			verify(forwarder).sendRepresentations(representations);
		}

		@Test
		void shouldCompleteRequest() {
			forwarder.forward(grpcRouteForwarding, attachments, representations);

			verify(requestObserver).onCompleted();
		}

		@Test
		void shouldWaitForCompletion() {
			forwarder.forward(grpcRouteForwarding, attachments, representations);

			verify(forwarder).waitForCompletion(future);
		}
	}

	@Nested
	class TestPerformGrpcCall {

		@BeforeEach
		void init() {
			when(serviceStub.withInterceptors(any())).thenReturn(serviceStub);
		}

		@Test
		void shouldAttachClientCallContextToServiceStub() {
			forwarder.performGrpcCall();

			verify(serviceStub).withInterceptors(any(VorgangManagerClientCallContextAttachingInterceptor.class));
		}

		@Test
		void shouldCreateResponseObserver() {
			forwarder.performGrpcCall();

			assertThat(getResponseObserverFromForwarder()).isNotNull();
		}

		@Test
		void shouldMakeGrpcCallToRouteForwarding() {
			forwarder.performGrpcCall();

			verify(serviceStub).routeForwarding(getResponseObserverFromForwarder());
		}

		@Test
		void shouldReturnFutureOfResponseObserver() {
			var result = forwarder.performGrpcCall();

			var expectedFuture = ReflectionTestUtils.getField(getResponseObserverFromForwarder(), "future", CompletableFuture.class);
			assertThat(result).isSameAs(expectedFuture);
		}
	}

	@Nested
	class TestSendRouteForwarding {

		private final GrpcRouteForwarding grpcRouteForwarding = GrpcRouteForwarding.newBuilder().build();
		@Mock
		private ForwardingResponseObserver responseObserver;
		@Mock
		private Runnable onReadyHandler;
		@Captor
		private ArgumentCaptor<Runnable> onReadyHandlerCaptor;
		@Captor
		private ArgumentCaptor<CompletableFuture<Void>> futureCaptor;

		@BeforeEach
		void init() {
			setResponseObserverInForwarder(responseObserver);
			doReturn(onReadyHandler).when(forwarder).getSendRouteForwardingRunnable(any(), any());
			doNothing().when(forwarder).waitForCompletion(any());
		}

		@SuppressWarnings("unchecked")
		@Test
		void shouldGetSendRouteForwardingRunnable() {
			forwarder.sendRouteForwarding(grpcRouteForwarding);

			verify(forwarder).getSendRouteForwardingRunnable(eq(grpcRouteForwarding), any(CompletableFuture.class));
		}

		@Test
		void shouldRegisterOnReadyHandler() {
			forwarder.sendRouteForwarding(grpcRouteForwarding);

			verify(responseObserver).registerOnReadyHandler(onReadyHandlerCaptor.capture());
			assertIsSendRouteForwardingRunnable(onReadyHandlerCaptor.getValue());
		}

		@SuppressWarnings("unchecked")
		@Test
		void shouldWaitForCompletion() {
			forwarder.sendRouteForwarding(grpcRouteForwarding);

			verify(forwarder).waitForCompletion(any(CompletableFuture.class));
		}

		@Test
		void shouldBeTheSameFuture() {
			forwarder.sendRouteForwarding(grpcRouteForwarding);

			verify(forwarder).getSendRouteForwardingRunnable(any(), futureCaptor.capture());
			verify(forwarder).waitForCompletion(futureCaptor.getValue());
			assertThat(futureCaptor.getAllValues().getFirst()).isSameAs(futureCaptor.getValue());
		}

		private void assertIsSendRouteForwardingRunnable(Runnable runnable) {
			runnable.run();
			verify(onReadyHandler).run();
		}
	}

	@Nested
	class TestGetSendRouteForwardingRunnable {

		private final GrpcRouteForwarding grpcRouteForwarding = GrpcRouteForwardingTestFactory.create();
		@Mock
		private ClientCallStreamObserver<GrpcRouteForwardingRequest> requestObserver;
		@Mock
		private CompletableFuture<Void> future;

		@BeforeEach
		void init() {
			setRequestObserverInForwarder(requestObserver);
		}

		@Test
		void shouldCallOnNext() {
			forwarder.getSendRouteForwardingRunnable(grpcRouteForwarding, future).run();

			verify(requestObserver).onNext(GrpcRouteForwardingRequestTestFactory.create());
		}

		@Test
		void shouldCompleteFuture() {
			forwarder.getSendRouteForwardingRunnable(grpcRouteForwarding, future).run();

			verify(future).complete(null);
		}

		@Test
		void shouldRunOnlyOnce() {
			var runnable = forwarder.getSendRouteForwardingRunnable(grpcRouteForwarding, future);

			IntStream.range(0, 3).forEach(i -> runnable.run());

			verify(requestObserver, times(1)).onNext(any());
		}
	}

	@Nested
	class TestSendAttachments {

		private final List<IncomingFileGroup> attachments = List.of(IncomingFileGroupTestFactory.createWithTwoFiles());

		@BeforeEach
		void init() {
			doNothing().when(forwarder).sendAttachmentFile(any());
		}

		@Test
		void shouldSendFirstAttachmentFile() {
			forwarder.sendAttachments(attachments);

			verify(forwarder).sendAttachmentFile(
					new EingangForwarder.FileInGroup(IncomingFileGroupTestFactory.NAME, IncomingFileGroupTestFactory.FILE));
		}

		@Test
		void shouldSendSecondAttachmentFile() {
			forwarder.sendAttachments(attachments);

			verify(forwarder).sendAttachmentFile(
					new EingangForwarder.FileInGroup(IncomingFileGroupTestFactory.NAME, IncomingFileGroupTestFactory.FILE2));
		}
	}

	@Nested
	class TestSendAttachmentFile {

		@Mock
		private StreamingFileSender<GrpcRouteForwardingRequest, GrpcRouteForwardingResponse> fileSender;
		@Mock
		private InputStream fileContentStream;

		@BeforeEach
		void init() {
			when(fileService.getUploadedFileStream(any())).thenReturn(fileContentStream);
			doReturn(fileSender).when(forwarder).createAttachmentFileSender(any(), any(), any());
			doReturn(fileSender).when(fileSender).send();
			doNothing().when(forwarder).waitForCompletion(any(), any());
		}

		@Test
		void shouldGetUploadFileStream() {
			sendAttachmentFile();

			verify(fileService).getUploadedFileStream(IncomingFileGroupTestFactory.FILE.getId());
		}

		@Test
		void shouldCreateAttachmentFileSender() {
			sendAttachmentFile();

			verify(forwarder).createAttachmentFileSender(IncomingFileGroupTestFactory.NAME, IncomingFileGroupTestFactory.FILE, fileContentStream);
		}

		@Test
		void shouldSend() {
			sendAttachmentFile();

			verify(fileSender).send();
		}

		@Test
		void shouldWaitForCompletion() {
			sendAttachmentFile();

			verify(forwarder).waitForCompletion(fileSender, fileContentStream);
		}

		private void sendAttachmentFile() {
			forwarder.sendAttachmentFile(new EingangForwarder.FileInGroup(IncomingFileGroupTestFactory.NAME, IncomingFileGroupTestFactory.FILE));
		}
	}

	@Nested
	class TestCreateAttachmentFileSender {

		@Mock
		private InputStream inputStream;
		@Mock
		private StreamingFileSender<GrpcRouteForwardingRequest, GrpcRouteForwardingResponse> fileSender;
		@Mock
		private StreamingFileSender<GrpcRouteForwardingRequest, GrpcRouteForwardingResponse> fileSenderWithMetadata;
		@Captor
		private ArgumentCaptor<BiFunction<byte[], Integer, GrpcRouteForwardingRequest>> chunkBuilderCaptor;

		private final byte[] chunk = RandomUtils.insecure().randomBytes(5);
		private final GrpcRouteForwardingRequest metadataRequest = GrpcRouteForwardingRequestTestFactory.create();

		@BeforeEach
		void init() {
			doReturn(fileSender).when(forwarder).createSenderWithoutMetadata(any(), any());
			doReturn(metadataRequest).when(forwarder).buildGrpcAttachmentFile(any(), any());
			when(fileSender.withMetaData(any())).thenReturn(fileSenderWithMetadata);
		}

		@Test
		void shouldCreateSenderWithoutMetadata() {
			createAttachmentFileSender();

			verify(forwarder).createSenderWithoutMetadata(chunkBuilderCaptor.capture(), eq(inputStream));
			chunkBuilderCaptor.getValue().apply(chunk, chunk.length);
			verify(forwarder).buildAttachmentChunk(chunk, chunk.length);
		}

		@Test
		void shouldBuildGrpcAttachmentFile() {
			createAttachmentFileSender();

			verify(forwarder).buildGrpcAttachmentFile(IncomingFileGroupTestFactory.NAME, IncomingFileGroupTestFactory.FILE);
		}

		@Test
		void shouldSetMetaData() {
			createAttachmentFileSender();

			verify(fileSender).withMetaData(metadataRequest);
		}

		@Test
		void shouldReturnBuiltFileSender() {
			var returnedFileSender = createAttachmentFileSender();

			assertThat(returnedFileSender).isSameAs(fileSenderWithMetadata);
		}

		private StreamingFileSender<GrpcRouteForwardingRequest, GrpcRouteForwardingResponse> createAttachmentFileSender() {
			return forwarder.createAttachmentFileSender(IncomingFileGroupTestFactory.NAME, IncomingFileGroupTestFactory.FILE, inputStream);
		}
	}

	@Nested
	class TestBuildAttachmentChunk {

		private final byte[] chunk = RandomUtils.insecure().randomBytes(5);

		@BeforeEach
		void mock() {
			doReturn(GrpcAttachmentTestFactory.CONTENT).when(forwarder).buildGrpcFileContent(any(), anyInt());
		}

		@Test
		void shouldCallBuildGrpcFileContent() {
			forwarder.buildAttachmentChunk(chunk, chunk.length);

			verify(forwarder).buildGrpcFileContent(chunk, chunk.length);
		}

		@Test
		void shouldReturnGrpcRouteForwardingRequest() {
			var result = forwarder.buildAttachmentChunk(chunk, chunk.length);

			assertThat(result).isEqualTo(GrpcRouteForwardingRequestTestFactory.createWithAttachmentContent());
		}
	}

	@Nested
	class TestBuildGrpcAttachmentFile {

		private final IncomingFile file = IncomingFileTestFactory.create();

		@BeforeEach
		void mock() {
			when(incomingFileMapper.toAttachmentFile(any(), any())).thenReturn(GrpcAttachmentFileTestFactory.create());
		}

		@Test
		void shouldCallIncomingFileMapper() {
			forwarder.buildGrpcAttachmentFile(IncomingFileGroupTestFactory.NAME, file);

			verify(incomingFileMapper).toAttachmentFile(IncomingFileGroupTestFactory.NAME, file);
		}

		@Test
		void shouldReturnAttachmentMetadataRequest() {
			var result = forwarder.buildGrpcAttachmentFile(IncomingFileGroupTestFactory.NAME, file);

			assertThat(result).isEqualTo(GrpcRouteForwardingRequestTestFactory.createWithAttachmentMetadata());
		}
	}

	@Nested
	class TestSendRepresentations {

		private static final IncomingFile FILE = IncomingFileTestFactory.create();

		@BeforeEach
		void init() {
			doNothing().when(forwarder).sendRepresentationFile(any());
		}

		@Test
		void shouldSendRepresentationFile() {
			forwarder.sendRepresentations(List.of(FILE));

			verify(forwarder).sendRepresentationFile(FILE);
		}
	}

	@Nested
	class TestSendRepresentationFile {

		@Mock
		private StreamingFileSender<GrpcRouteForwardingRequest, GrpcRouteForwardingResponse> fileSender;
		@Mock
		private InputStream fileContentStream;

		private final IncomingFile file = IncomingFileTestFactory.create();

		@BeforeEach
		void init() {
			when(fileService.getUploadedFileStream(any())).thenReturn(fileContentStream);
			doReturn(fileSender).when(forwarder).createRepresentationFileSender(any(), any());
			doReturn(fileSender).when(fileSender).send();
			doNothing().when(forwarder).waitForCompletion(any(), any());
		}

		@Test
		void shouldGetUploadFileStream() {
			sendRepresentationFile();

			verify(fileService).getUploadedFileStream(IncomingFileTestFactory.ID);
		}

		@Test
		void shouldCreateRepresentationFileSender() {
			sendRepresentationFile();

			verify(forwarder).createRepresentationFileSender(file, fileContentStream);
		}

		@Test
		void shouldSend() {
			sendRepresentationFile();

			verify(fileSender).send();
		}

		@Test
		void shouldWaitForCompletion() {
			sendRepresentationFile();

			verify(forwarder).waitForCompletion(fileSender, fileContentStream);
		}

		private void sendRepresentationFile() {
			forwarder.sendRepresentationFile(file);
		}
	}

	@Nested
	class TestCreateRepresentationFileSender {

		@Mock
		private InputStream inputStream;
		@Mock
		private StreamingFileSender<GrpcRouteForwardingRequest, GrpcRouteForwardingResponse> fileSender;
		@Mock
		private StreamingFileSender<GrpcRouteForwardingRequest, GrpcRouteForwardingResponse> fileSenderWithMetadata;
		@Captor
		private ArgumentCaptor<BiFunction<byte[], Integer, GrpcRouteForwardingRequest>> chunkBuilderCaptor;

		private final byte[] chunk = RandomUtils.insecure().randomBytes(5);
		private final GrpcRouteForwardingRequest metadataRequest = GrpcRouteForwardingRequestTestFactory.create();
		private final IncomingFile incomingFile = IncomingFileTestFactory.create();

		@BeforeEach
		void init() {
			doReturn(fileSender).when(forwarder).createSenderWithoutMetadata(any(), any());
			doReturn(metadataRequest).when(forwarder).buildGrpcRepresentationFile(any());
			when(fileSender.withMetaData(any())).thenReturn(fileSenderWithMetadata);
		}

		@Test
		void shouldCallCreateSenderWithoutMetadata() {
			createRepresentationFileSender();

			verify(forwarder).createSenderWithoutMetadata(chunkBuilderCaptor.capture(), eq(inputStream));
			chunkBuilderCaptor.getValue().apply(chunk, chunk.length);
			verify(forwarder).buildRepresentationChunk(chunk, chunk.length);
		}

		@Test
		void shouldCallBuildGrpcRepresentationFile() {
			createRepresentationFileSender();

			verify(forwarder).buildGrpcRepresentationFile(incomingFile);
		}

		@Test
		void shouldSetMetaData() {
			createRepresentationFileSender();

			verify(fileSender).withMetaData(metadataRequest);
		}

		@Test
		void shouldReturnBuiltFileSender() {
			var returnedFileSender = createRepresentationFileSender();

			assertThat(returnedFileSender).isSameAs(fileSenderWithMetadata);
		}

		private StreamingFileSender<GrpcRouteForwardingRequest, GrpcRouteForwardingResponse> createRepresentationFileSender() {
			return forwarder.createRepresentationFileSender(incomingFile, inputStream);
		}
	}

	@Nested
	class TestBuildRepresentationChunk {

		private final byte[] chunk = RandomUtils.insecure().randomBytes(5);

		@BeforeEach
		void mock() {
			doReturn(GrpcRepresentationTestFactory.CONTENT).when(forwarder).buildGrpcFileContent(any(), anyInt());
		}

		@Test
		void shouldCallBuildGrpcFileContent() {
			forwarder.buildRepresentationChunk(chunk, chunk.length);

			verify(forwarder).buildGrpcFileContent(chunk, chunk.length);
		}

		@Test
		void shouldReturnGrpcRouteForwardingRequest() {
			var result = forwarder.buildRepresentationChunk(chunk, chunk.length);

			assertThat(result).isEqualTo(GrpcRouteForwardingRequestTestFactory.createWithRepresentationContent());
		}
	}

	@Nested
	class TestBuildGrpcRepresentationFile {

		private final IncomingFile file = IncomingFileTestFactory.create();

		@BeforeEach
		void mock() {
			when(incomingFileMapper.toRepresentationFile(any())).thenReturn(GrpcRepresentationFileTestFactory.create());
		}

		@Test
		void shouldCallIncomingFileMapper() {
			forwarder.buildGrpcRepresentationFile(file);

			verify(incomingFileMapper).toRepresentationFile(file);
		}

		@Test
		void shouldReturnRepresentationMetadataRequest() {
			var result = forwarder.buildGrpcRepresentationFile(file);

			assertThat(result).isEqualTo(GrpcRouteForwardingRequestTestFactory.createWithRepresentationMetadata());
		}
	}

	@Nested
	class TestBuildGrpcFileContent {

		@Nested
		class TestOnEndOfFile {

			@Test
			void shouldBuildEndOfFileChunk() {
				var fileContent = forwarder.buildGrpcFileContent(new byte[0], -1);

				assertThat(fileContent).isEqualTo(GrpcFileContentTestFactory.createEndOfFile());
			}
		}

		@Nested
		class TestOnContentProvided {

			@Test
			void shouldBuildEndOfFileChunk() {
				var fileContent = forwarder.buildGrpcFileContent(GrpcFileContentTestFactory.CONTENT, GrpcFileContentTestFactory.CONTENT.length);

				assertThat(fileContent).isEqualTo(GrpcFileContentTestFactory.create());
			}
		}
	}

	@Nested
	class TestCreateSenderWithoutMetadata {

		private MockedStatic<GrpcFileUploadUtils> grpcFileUploadUtilsMock;
		@Mock
		private BiFunction<byte[], Integer, GrpcRouteForwardingRequest> chunkBuilder;
		@Mock
		private ClientCallStreamObserver<GrpcRouteForwardingRequest> requestObserver;
		@Mock
		private ForwardingResponseObserver responseObserver;
		@Mock
		private InputStream inputStream;
		@Mock
		private StreamingFileSender<GrpcRouteForwardingRequest, GrpcRouteForwardingResponse> fileSender;
		@Mock
		private Runnable onReadyHandler;
		@Captor
		private ArgumentCaptor<Consumer<Runnable>> onReadyHandlerCaptor;

		@BeforeEach
		void init() {
			grpcFileUploadUtilsMock = mockStatic(GrpcFileUploadUtils.class);
			grpcFileUploadUtilsMock.when(() -> GrpcFileUploadUtils.createStreamSharingSender(any(), any(), any(), any())).thenReturn(fileSender);
			ReflectionTestUtils.setField(forwarder, "responseObserver", responseObserver);
			ReflectionTestUtils.setField(forwarder, "requestObserver", requestObserver);
		}

		@AfterEach
		void tearDown() {
			grpcFileUploadUtilsMock.close();
		}

		@Test
		void shouldCreateFileSender() {

			createSenderWithoutMetadata();

			grpcFileUploadUtilsMock
					.verify(() -> GrpcFileUploadUtils.<GrpcRouteForwardingRequest, GrpcRouteForwardingResponse>createStreamSharingSender(
							eq(chunkBuilder), eq(inputStream), eq(requestObserver), onReadyHandlerCaptor.capture()));
			assertIsRegisterOnReadyHandler(onReadyHandlerCaptor);
		}

		@Test
		void shouldReturnCreatedFileSender() {
			var returnedFileSender = createSenderWithoutMetadata();

			assertThat(returnedFileSender).isSameAs(fileSender);
		}

		private void assertIsRegisterOnReadyHandler(ArgumentCaptor<Consumer<Runnable>> captor) {
			captor.getValue().accept(onReadyHandler);
			verify(responseObserver).registerOnReadyHandler(onReadyHandler);
		}

		private StreamingFileSender<GrpcRouteForwardingRequest, GrpcRouteForwardingResponse> createSenderWithoutMetadata() {
			return forwarder.createSenderWithoutMetadata(chunkBuilder, inputStream);
		}
	}

	@Nested
	class TestWaitForCompletionOfFuture {

		@Mock
		private CompletableFuture<Void> future;

		@SneakyThrows
		@Test
		void shouldGetFromFuture() {
			waitForCompletion();

			verify(future).get(EingangForwarder.TIMEOUT_MINUTES, TimeUnit.MINUTES);
		}

		@Nested
		class TestOnInterruptedException {

			private final InterruptedException exception = new InterruptedException();

			@BeforeEach
			@SneakyThrows
			void mock() {
				when(future.get(anyLong(), any())).thenThrow(exception);
			}

			@Test
			void shouldThrowTechnicalException() {
				assertThrows(TechnicalException.class, TestWaitForCompletionOfFuture.this::waitForCompletion);
			}

			@Test
			void shouldInterruptThread() {
				try {
					waitForCompletion();
				} catch (TechnicalException e) {
					// expected
				}

				assertThat(Thread.currentThread().isInterrupted()).isTrue();
			}
		}

		@Nested
		class TestOnExecutionException {

			private final ExecutionException exception = new ExecutionException(new Exception());

			@BeforeEach
			@SneakyThrows
			void mock() {
				when(future.get(anyLong(), any())).thenThrow(exception);
			}

			@Test
			void shouldThrowTechnicalException() {
				assertThrows(TechnicalException.class, TestWaitForCompletionOfFuture.this::waitForCompletion);
			}
		}

		@Nested
		class TestOnTimeoutException {

			private final TimeoutException exception = new TimeoutException();

			@BeforeEach
			@SneakyThrows
			void mock() {
				when(future.get(anyLong(), any())).thenThrow(exception);
			}

			@Test
			void shouldThrowTechnicalException() {
				assertThrows(TechnicalException.class, TestWaitForCompletionOfFuture.this::waitForCompletion);
			}
		}

		private void waitForCompletion() {
			forwarder.waitForCompletion(future);
		}
	}

	@Nested
	class TestWaitForCompletionOfFileSender {

		@Mock
		private StreamingFileSender<GrpcRouteForwardingRequest, GrpcRouteForwardingResponse> fileSender;
		@Mock
		private InputStream fileContentStream;
		@Mock
		private CompletableFuture<GrpcRouteForwardingResponse> future;

		@BeforeEach
		void init() {
			when(fileSender.getResultFuture()).thenReturn(future);
		}

		@SneakyThrows
		@Test
		void shouldGetFromFuture() {
			waitForCompletion();

			verify(future).get(EingangForwarder.TIMEOUT_MINUTES, TimeUnit.MINUTES);
		}

		@Nested
		class TestOnInterruptedException {

			private final InterruptedException exception = new InterruptedException();

			@BeforeEach
			@SneakyThrows
			void mock() {
				when(future.get(anyLong(), any())).thenThrow(exception);
			}

			@Test
			void shouldInterruptThread() {
				try {
					waitForCompletion();
				} catch (TechnicalException e) {
					// expected
				}

				assertThat(Thread.currentThread().isInterrupted()).isTrue();
			}

			@Test
			void shouldCancelOnError() {
				try {
					waitForCompletion();
				} catch (TechnicalException e) {
					// expected
				}

				verify(fileSender).cancelOnError(exception);
			}

			@Test
			void shouldThrowTechnicalException() {
				assertThrows(TechnicalException.class, TestWaitForCompletionOfFileSender.this::waitForCompletion);
			}

			@SneakyThrows
			@Test
			void shouldCloseStream() {
				try {
					waitForCompletion();
				} catch (TechnicalException e) {
					// expected
				}

				verify(fileContentStream).close();
			}
		}

		@Nested
		class TestOnExecutionException {

			private final ExecutionException exception = new ExecutionException(new Exception());

			@BeforeEach
			@SneakyThrows
			void mock() {
				when(future.get(anyLong(), any())).thenThrow(exception);
			}

			@Test
			void shouldCancelOnError() {
				try {
					waitForCompletion();
				} catch (TechnicalException e) {
					// expected
				}

				verify(fileSender).cancelOnError(exception);
			}

			@Test
			void shouldThrowTechnicalException() {
				assertThrows(TechnicalException.class, TestWaitForCompletionOfFileSender.this::waitForCompletion);
			}

			@SneakyThrows
			@Test
			void shouldCloseStream() {
				try {
					waitForCompletion();
				} catch (TechnicalException e) {
					// expected
				}

				verify(fileContentStream).close();
			}
		}

		@Nested
		class TestOnTimeoutException {

			private final TimeoutException exception = new TimeoutException();

			@BeforeEach
			@SneakyThrows
			void mock() {
				when(future.get(anyLong(), any())).thenThrow(exception);
			}

			@Test
			void shouldCancelOnTimeout() {
				try {
					waitForCompletion();
				} catch (TechnicalException e) {
					// expected
				}

				verify(fileSender).cancelOnTimeout();
			}

			@Test
			void shouldThrowTechnicalException() {
				assertThrows(TechnicalException.class, TestWaitForCompletionOfFileSender.this::waitForCompletion);
			}

			@SneakyThrows
			@Test
			void shouldCloseStream() {
				try {
					waitForCompletion();
				} catch (TechnicalException e) {
					// expected
				}

				verify(fileContentStream).close();
			}
		}

		private void waitForCompletion() {
			forwarder.waitForCompletion(fileSender, fileContentStream);
		}
	}

	@Nested
	class TestForwardingResponseObserver {

		@Mock
		private CompletableFuture<GrpcRouteForwardingResponse> future;
		@Mock
		private DelegatingOnReadyHandler onReadyHandler;
		@Mock
		private ClientCallStreamObserver<GrpcRouteForwardingRequest> requestStream;
		private final GrpcRouteForwardingResponse response = GrpcRouteForwardingResponse.getDefaultInstance();
		@InjectMocks
		private ForwardingResponseObserver observer;

		@Nested
		class TestBeforeStart {

			@Test
			void shouldCreateOnReadyHandler() {
				observer.beforeStart(requestStream);

				assertThat(getOnReadyHandlerFromObserver()).isNotNull();
			}

			@Test
			void shouldSetOnReadyHandler() {
				observer.beforeStart(requestStream);

				verify(requestStream).setOnReadyHandler(getOnReadyHandlerFromObserver());
			}
		}

		@Nested
		class TestOnNext {

			@Test
			void shouldSetResponse() {
				observer.onNext(response);

				assertThat(getResponseFromObserver()).isSameAs(response);
			}
		}

		@Nested
		class TestOnError {

			private final Throwable error = new RuntimeException("Error when forwarding");

			@BeforeEach
			void init() {
				setOnReadyHandlerInObserver();
			}

			@Test
			void shouldStopOnReadyHandler() {
				observer.onError(error);

				verify(onReadyHandler).stop();
			}

			@Test
			void shouldCompleteFutureExceptionally() {
				observer.onError(error);

				verify(future).completeExceptionally(error);
			}
		}

		@Nested
		class TestOnCompleted {

			@BeforeEach
			void init() {
				setOnReadyHandlerInObserver();
			}

			@Test
			void shouldStopOnReadyHandler() {
				observer.onCompleted();

				verify(onReadyHandler).stop();
			}

			@Test
			void shouldCompleteFutureWithResponse() {
				observer.onNext(response);

				observer.onCompleted();

				verify(future).complete(response);
			}
		}

		@Nested
		class TestRegisterOnReadyHandler {

			@Mock
			private Runnable delegate;

			@BeforeEach
			void init() {
				setOnReadyHandlerInObserver();
			}

			@Test
			void shouldSetDelegateInOnReadyHandler() {
				observer.registerOnReadyHandler(delegate);

				verify(onReadyHandler).setDelegate(delegate);
			}
		}

		private DelegatingOnReadyHandler getOnReadyHandlerFromObserver() {
			return ReflectionTestUtils.getField(observer, "onReadyHandler", DelegatingOnReadyHandler.class);
		}

		private void setOnReadyHandlerInObserver() {
			ReflectionTestUtils.setField(observer, "onReadyHandler", onReadyHandler);
		}

		private GrpcRouteForwardingResponse getResponseFromObserver() {
			return ReflectionTestUtils.getField(observer, "response", GrpcRouteForwardingResponse.class);
		}
	}

	@Nested
	class TestDelegatingOnReadyHandler {

		@Mock
		private ClientCallStreamObserver<GrpcRouteForwardingRequest> requestStream;
		@InjectMocks
		private DelegatingOnReadyHandler onReadyHandler;

		@Test
		void shouldDoneBeInitiallyFalse() {
			assertThat(getDoneFromOnReadyHandler()).isFalse();
		}

		@Nested
		class TestSetDelegate {

			@Mock
			private Runnable delegate;

			@Test
			void shouldSetDelegate() {
				onReadyHandler.setDelegate(delegate);

				assertThat(getDelegateFromOnReadyHandler()).isSameAs(delegate);
			}
		}

		@Nested
		class TestStop {

			@Test
			void shouldSetDoneToTrue() {
				onReadyHandler.stop();

				assertThat(getDoneFromOnReadyHandler()).isTrue();
			}
		}

		@Nested
		class TestRun {

			@Mock
			private Runnable delegate;

			@BeforeEach
			void init() {
				onReadyHandler.setDelegate(delegate);
			}

			@Test
			void shouldNotRunDelegateIfDone() {
				onReadyHandler.stop();
				lenient().when(requestStream.isReady()).thenReturn(true);

				onReadyHandler.run();

				verify(delegate, never()).run();
			}

			@Test
			void shouldNotRunDelegateIfNotReady() {
				when(requestStream.isReady()).thenReturn(false);

				onReadyHandler.run();

				verify(delegate, never()).run();
			}

			@Test
			void shouldRunDelegateIfNotDoneAndReady() {
				when(requestStream.isReady()).thenReturn(true).thenReturn(false);
				runWithOnReadyHandlerInAnotherThread(() -> {
					await().atMost(Duration.ofMillis(500)).untilAsserted(() -> verify(delegate, atLeastOnce()).run());
				});
			}

			@Test
			void shouldContinueAfterDelegateWasReplaced() {
				when(requestStream.isReady()).thenReturn(true);
				runWithOnReadyHandlerInAnotherThread(() -> {
					await().atMost(Duration.ofMillis(500)).untilAsserted(() -> verify(delegate, atLeastOnce()).run());
					var delegate2 = mock(Runnable.class);
					onReadyHandler.setDelegate(delegate2);
					await().atMost(Duration.ofMillis(500)).untilAsserted(() -> verify(delegate2, atLeastOnce()).run());
				});
			}

			private void runWithOnReadyHandlerInAnotherThread(Runnable runnable) {
				try (ExecutorService executor = Executors.newSingleThreadExecutor()) {
					var future = executor.submit(onReadyHandler);
					runnable.run();
					future.cancel(true);
					executor.shutdown();
				}
			}
		}

		private boolean getDoneFromOnReadyHandler() {
			return ReflectionTestUtils.getField(onReadyHandler, "done", AtomicBoolean.class).get();
		}

		private Runnable getDelegateFromOnReadyHandler() {
			return (Runnable) ReflectionTestUtils.getField(onReadyHandler, "delegateRef", AtomicReference.class).get();
		}
	}

	private void setResponseObserverInForwarder(ForwardingResponseObserver responseObserver) {
		ReflectionTestUtils.setField(forwarder, "responseObserver", responseObserver);
	}

	private ForwardingResponseObserver getResponseObserverFromForwarder() {
		return ReflectionTestUtils.getField(forwarder, "responseObserver", ForwardingResponseObserver.class);
	}

	private void setRequestObserverInForwarder(ClientCallStreamObserver<GrpcRouteForwardingRequest> requestObserver) {
		ReflectionTestUtils.setField(forwarder, "requestObserver", requestObserver);
	}
}
